@@ -32,6 +32,11 @@ from pycollector.util import allmondays_since, yyyymmdd_to_date, is_email_addres
 from pycollector.util import lowerif, timestamp, fromdate, ismonday
 from pycollector.globals import print
 
+try:
+    import ujson as json  # faster
+except ImportError:
+    import json
+
 
 class Video(Scene):
     """pycollector.video.Video class
@@ -74,13 +79,40 @@ class Video(Scene):
         if fetch:
             self._load_json()
 
+    @classmethod
+    def from_json(obj, s):
+        d = json.loads(s) if not isinstance(s, dict) else s
+        v = Scene.from_json(d)  
+        v._is_json_loaded = d['_is_json_loaded']
+        v._dt = d['_dt']
+        v._mindim = d['_mindim']
+        v._verbose = d['_verbose']
+        v._jsonfile = d['_jsonfile']
+        v._jsonurl = d['_jsonurl']        
+        v._mp4file = d['_mp4file']
+        v._mp4url = d['_mp4url']
+        v.__class__ = Video
+        return v
+        
+    def json(self, encode=True):
+        d = super().json(encode=False)
+        d['_is_json_loaded'] = self._is_json_loaded
+        d['_dt'] = self._dt
+        d['_mindim'] = self._mindim
+        d['_verbose'] = self._verbose
+        d['_jsonfile'] = self._jsonfile
+        d['_jsonurl'] = self._jsonurl        
+        d['_mp4file'] = self._mp4file
+        d['_mp4url'] = self._mp4url
+        return json.dumps(d) if encode else d
+    
     def __repr__(self):
         return str("<pycollector.video: uploaded=%s, activities=%s, scene=%s>" % (self.timestamp().strftime("%Y-%m-%d %H:%M")
                                                                                   if self.timestamp() is not None
                                                                                   else str(None),
                                                                                   self.activity_categories(),
                                                                                   str(super().__repr__())))
-            
+    
     def _load_json(self):
         """Lazy JSON download, parse, and import"""
 
@@ -153,22 +185,58 @@ class Video(Scene):
                 
 
         else:
-            print('[pycollector.video]: empty JSON "%s" - SKIPPING' % jsonfile)
+            print('[pycollector.video]: empty JSON "%s"' % jsonfile)
             d = None
 
         # Backwards compatible video import: should not be necessary with new app release
-        if not 'category' in d['metadata']:
+        if d is not None and not 'category' in d['metadata']:
             vipy.util.try_import('pycollector.admin.globals', message="Not authorized - Old style JSON requires admin access")
-            from pycollector.admin.globals import backend  
-            if not backend().collections().iscollectionid(d["metadata"]["collection_id"]):
-                print('[pycollector.video]: invalid collection ID "%s" - SKIPPING' % d["metadata"]["collection_id"])
-                d = None
+            from pycollector.admin.globals import backend, isapi
+            from pycollector.admin.legacy import applabel_to_longlabel, shortname_synonyms, applabel_to_piplabel
+
+            # V1 - old collection name pattern
+            if any([d["metadata"]["collection_id"] in k for k in applabel_to_piplabel().keys()]):
+                try:
+                    d['metadata']['collection_name'] = d["metadata"]["collection_id"]
+                    applabel = ['%s_%s_%s' % (d["metadata"]["project_id"], d["metadata"]["collection_id"], a['label']) for a in d['activity']]
+                    applabel = [a if a in applabel_to_piplabel() else '%s_%s_%s' % (d["metadata"]["project_id"], d["metadata"]["collection_id"], shortname_synonyms()[a.split('_')[2]]) for a in applabel]
+                    d['metadata']['category'] = ','.join([applabel_to_piplabel()[a] for a in applabel])
+                    d['metadata']['shortname'] = ','.join([a.split('_')[2] for a in applabel])
+                except Exception as e:
+                    print('[pycollector.video]: legacy json import failed for v1 JSON "%s" with error "%s"' % (str(d['metadata']), str(e)))
+                    d = None
+
+            # V2 - new collection names, but activity names not in JSON
+            elif isapi('v1') or isapi('v2'):
+                version = 'v1' if isapi('v1') else 'v2'
+                if version == 'v1':
+                    backend('prod', 'v2')  # temporary switch
+
+                if not backend().collections().iscollectionid(d["metadata"]["collection_id"]):
+                    print('[pycollector.video]: invalid collection ID "%s"' % d["metadata"]["collection_id"])
+                    d = None
+                else:
+                    try:
+                        # Fetch labels from backend (with legacy shortname translation)
+                        C = backend().collections()[d["metadata"]["collection_id"]]
+                        d['metadata']['collection_name'] = backend().collections().id_to_name(d["metadata"]["collection_id"])
+                        shortnames = []
+                        for a in d['activity']:
+                            if not (a['label'] in C.shortnames() or a['label'] in shortname_synonyms()):
+                                raise ValueError("Invalid shortname '%s' for collection shortnames '%s' and not in legacy synonyms '%s'" % (a['label'], str(C.shortnames()), str(shortname_synonyms())))
+                            shortnames.append(a['label'] if a['label'] in C.shortnames() else shortname_synonyms()[a['label']])
+                        d['metadata']['category'] = ','.join([C.shortname_to_activity(s, strict=False) for s in shortnames])
+                        d['metadata']['shortname'] = ','.join([s for s in shortnames])
+                    except Exception as e: 
+                        print('[pycollector.video]: label fetch failed for %s with exception %s' % (str(d['activity']), str(e)))
+                        d = None
+
+                if version == 'v1':
+                    backend('prod', 'v1')  # switch back
             else:
-                # Fetch labels from backend (yuck)
-                d['metadata']['collection_name'] = backend().collection().id_to_name(d["metadata"]["collection_id"])
-                d['metadata']['category'] = ','.join([backend().collection()[d["metadata"]["collection_id"]].shortname_to_activity(a["label"]) for a in d['activity']])
-                d['metadata']['shortname'] = ','.join([a["label"] for a in d['activity']])
-                
+                print('[pycollector.video]: Legacy JSON import failed for JSON with metadata - "%s"' % str(d["metadata"]))
+                d = None
+            
         else:
             # New style JSON: use labels stored directly in JSON
             pass
@@ -218,7 +286,7 @@ class Video(Scene):
                 t = Track(
                     category=obj["label"],
                     framerate=float(d["metadata"]["frame_rate"]),
-                    keyframes=[f for (f, bb) in zip(keyframes, keyboxes) if bb.isvalid()],
+                    keyframes=[int(f) for (f, bb) in zip(keyframes, keyboxes) if bb.isvalid()],
                     boxes=[bb for (f, bb) in zip(keyframes, keyboxes) if bb.isvalid()],
                     boundary="strict",
                 )
@@ -247,6 +315,11 @@ class Video(Scene):
             # Import activities
             for a in d["activity"]:
                 try:
+                    # Legacy shortname display
+                    if a['label'] not in d_shortname_to_category:
+                        if a['label'] not in shortname_synonyms():
+                            raise ValueError("Invalid shortname '%s' for collection shortnames '%s' and not in legacy synonyms '%s'" % (a['label'], d_shortname_to_category, str(shortname_synonyms())))
+                        a['label'] = a['label'] if a['label'] in d_shortname_to_category else shortname_synonyms()[a['label']]  # legacy translation              
                     if (d["metadata"]["collection_id"] == "P004C009" and d["metadata"]["device_identifier"] == "android"):
                         shortlabel = "Buying (Machine)"
                     elif (d["metadata"]["collection_id"] == "P004C008" and d["metadata"]["device_identifier"] == "ios" and "Purchasing" in a["label"]):
@@ -282,7 +355,7 @@ class Video(Scene):
                                             
                 except Exception as e:
                     print(
-                        '[pycollector.video]: Filtering invalid activity "%s" with error "%s" for videoid=%s'
+                        '[pycollector.video]: Filtering invalid activity JSON "%s" with error "%s" for videoid=%s'
                         % (str(a), str(e), d["metadata"]["video_id"])
                     )
 
@@ -307,12 +380,15 @@ class Video(Scene):
                     self.clear()  # remove this old video from consideration
                     self._is_json_loaded = False
         else:
+            print('[pycollector.video]: JSON load failed - SKIPPING')
             self._is_json_loaded = False
 
         # Resample tracks
         if self._dt > 1 and self._is_json_loaded:
-            self.trackmap(lambda t: t.resample(self._dt))
-        
+            self.trackmap(lambda t: t.resample(self._dt).significant_digits(2))
+
+        assert vipy.version.is_at_least('1.8.34')            
+        self.trackmap(lambda t: t.significant_digits(2))            
         return self
 
     def variant(self):
@@ -426,6 +502,9 @@ class Video(Scene):
     def collectorid(self):
         return self.attributes["collector_id"] if "collector_id" in self._load_json().attributes else None
 
+    def subjectid(self):
+        return self.attributes["subject_id"][0] if "subject_id" in self._load_json().attributes else None
+        
     def collectionid(self):
         return self.attributes["collection_id"] if "collection_id" in self._load_json().attributes else None
 
@@ -462,12 +541,32 @@ class Video(Scene):
         v.__class__ = pycollector.admin.video.Video
         return v
         
+    def project(self):
+        return self.attributes['project_name']
+
+    def program(self):
+        return self.attributes['program_name']
+
+    def objectdetection(self, frame=1):
+        """Run an object detector on a given frame of video.  It is more efficient to construct an ObjectDetector() object once and reuse it."""        
+        from pycollector.detection import ObjectDetector
+        return ObjectDetector()(self.frame(frame))
+
+    def facedetection(self, frame=1):
+        """Run face detection on a given frame of video.  It is more efficient to construct a FaceDetector() object once and reuse it."""
+        from pycollector.detection import FaceDetector        
+        return FaceDetector()(self.frame(frame))
+    
+    
 def search():
     import pycollector.project
     return pycollector.project.Project(since='2020-09-01')
 
 
-def last(n=1):
+def last(n=1, program=None):
     import pycollector.project    
-    return pycollector.project.Project(since='2020-09-01', last=n).last(n)
 
+    if program:
+        return pycollector.project.Project(program_id=program, since='2020-09-01', last=n).last(n)
+    else:
+        return pycollector.project.Project(since='2020-09-01', last=n).last(n)
